* SLHARead.F
* read a file in SUSY Les Houches Accord format (hep-ph/0311123)
* last modified 30 Aug 04 th

#include "PDG.h"
#include "SLHAData.h"


	subroutine SLHARead(error, slhadata, filename, abort)
	implicit none
	integer error, abort
	double precision slhadata(*)
	character*(*) filename

	integer lnblnk, SLHATokens
	external lnblnk, SLHATokens

	integer lineno, block, n, n1, n2, pdgno
	integer decay, parent, parent_id, i, id(10)
	double precision blockQ, val

	integer mask
	parameter (mask = 2**30 - 1)

#define Range(i,min,max) iand(max-i,mask).le.max-min
#define Var(x,n,min,max) if( Range(n,min,max) ) x(n)
#define Array(x,n1,n2,max) if( Range(n1,1,max) .and. Range(n2,1,max) ) x(n1,n2)

	character*200 line

	integer nblocks
	parameter (nblocks = 23)
	character*10 blockname(nblocks)
	data blockname / "MODSEL", "SMINPUTS", "MINPAR", "EXTPAR",
     &    "MASS", "NMIX", "UMIX", "VMIX",
     &    "STAUMIX", "STOPMIX", "SBOTMIX",
     &    "ALPHA", "HMIX", "GAUGE", "MSOFT",
     &    "AE", "AU", "AD", "YE", "YU", "YD", "SPINFO", "DCINFO" /

	error = 0

	open(10, file=filename, status="old", err=97)
	lineno = 0
	block = 0
	decay = 1
	pdgno = 0

100	format(A)
101	format(A, I5)
102	format(A, A, A, I5)
103	format(A, A)

7	continue
	lineno = lineno + 1
	read(10, 100, end=99) line
	if( SLHATokens(line) .eq. 0 ) goto 7

	if( line(1:6) .eq. "BLOCK " ) then
	  n = index(line(7:), " ")
	  do block = 1, nblocks
	    if( line(7:n+6) .eq. blockname(block)(1:n) ) then
	      blockQ = 0
	      if( line(n+7:n+8) .eq. "Q=" ) read(line(n+9:), *) blockQ
	      goto 7
	    endif
	  enddo
	  print 102, "Unknown block ", line(7:n+5), " in input line", lineno
	  block = -1
	  goto 7
	endif

	if( line(1:6) .eq. "DECAY " ) then
	  if( decay + 2 .ge. LengthDecay ) then
	    print 101, "Out of memory in input line", lineno
	    block = -1
	  else
	    block = nblocks + 1
	    read(line(7:), *, err=98, end=98) parent_id, val
	    parent = decay
	    decay = decay + 2
	    Decay(parent) = Decay_Entry(parent_id, 2)
	    Decay(parent + 1) = val
	  endif
	  goto 7
	endif

	goto (1001, 1002, 1003, 1004,
     &    2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008,
     &    2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
     &    2017, 2018, 2018, 3000) block

	if( block .eq. 0 ) then
	  if( abort .ne. 0 ) then
* depending on the abort flag, one can stop reading here and
* assume the file is in a different format
	    error = 2
	    goto 99
	  endif
* or issue a warning and continue
	  print 101, "Superfluous text in input line", lineno
	endif

	goto 7

97	print 103, "Cannot open ", filename
	error = 1
	return

98	print 101, "Syntax error in input line", lineno
	goto 7

* Block MODSEL
1001	read(line, *, err=98, end=98) n, val
	if( n .eq. 1  ) ModSel_Model = val
	if( n .eq. 3  ) ModSel_Content = val
	if( n .eq. 11 ) ModSel_GridPts = val
	if( n .eq. 12 ) ModSel_Qmax = val
	if( n .eq. 21 ) then
	  pdgno = pdgno + 1
	  if( pdgno .le. 5 ) ModSel_PDG(pdgno) = val
	endif
	goto 7

* Block SMINPUTS
1002	read(line, *, err=98, end=98) n, val
	if( n .eq. 1 ) SMInputs_AlfaMZ = val
	if( n .eq. 2 ) SMInputs_GF = val
	if( n .eq. 3 ) SMInputs_AlfasMZ = val
	if( n .eq. 4 ) SMInputs_MZ = val
	if( n .eq. 5 ) SMInputs_Mb = val
	if( n .eq. 6 ) SMInputs_Mt = val
	if( n .eq. 7 ) SMInputs_Mtau = val
	goto 7

* Block MINPAR
1003	read(line, *, err=98, end=98) n, val
	Var(BlockMinPar, n + 1, 1, 7) = val
	goto 7

* Block EXTPAR
1004	read(line, *, err=98, end=98) n, val
	if( n .eq. 0  ) ExtPar_Q = val
	if( n .eq. 1  ) ExtPar_M1 = val
	if( n .eq. 2  ) ExtPar_M2 = val
	if( n .eq. 3  ) ExtPar_M3 = val
	if( n .eq. 11 ) ExtPar_At = val
	if( n .eq. 12 ) ExtPar_Ab = val
	if( n .eq. 13 ) ExtPar_Atau = val
	if( n .eq. 21 ) ExtPar_MHd2 = val
	if( n .eq. 22 ) ExtPar_MHu2 = val
	if( n .eq. 23 ) ExtPar_MUE = val
	if( n .eq. 24 ) ExtPar_MA02 = val
	if( n .eq. 25 ) ExtPar_TB = val
	if( Range(n, 31, 33) ) ExtPar_MSL(n - 30) = val
	if( Range(n, 34, 36) ) ExtPar_MSE(n - 33) = val
	if( Range(n, 41, 43) ) ExtPar_MSQ(n - 40) = val
	if( Range(n, 44, 46) ) ExtPar_MSU(n - 43) = val
	if( Range(n, 47, 49) ) ExtPar_MSD(n - 46) = val
	if( Range(n, 51, 53) ) ExtPar_N5(n - 50) = val
	goto 7

* Block MASS
2001	read(line, *, err=98, end=98) n, val
	if( n .eq. PDG_nu_e        ) Mass_Mf(1,1) = val
	if( n .eq. PDG_electron    ) Mass_Mf(2,1) = val
	if( n .eq. PDG_up          ) Mass_Mf(3,1) = val
	if( n .eq. PDG_down        ) Mass_Mf(4,1) = val
	if( n .eq. PDG_nu_mu       ) Mass_Mf(1,2) = val
	if( n .eq. PDG_muon        ) Mass_Mf(2,2) = val
	if( n .eq. PDG_charm       ) Mass_Mf(3,2) = val
	if( n .eq. PDG_strange     ) Mass_Mf(4,2) = val
	if( n .eq. PDG_nu_tau      ) Mass_Mf(1,3) = val
	if( n .eq. PDG_tau         ) Mass_Mf(2,3) = val
	if( n .eq. PDG_top         ) Mass_Mf(3,3) = val
	if( n .eq. PDG_bottom      ) Mass_Mf(4,3) = val
	if( n .eq. PDG_snu_e1      ) Mass_MSf(1,1,1) = val
	if( n .eq. PDG_snu_e2      ) Mass_MSf(2,1,1) = val
	if( n .eq. PDG_selectron1  ) Mass_MSf(1,2,1) = val
	if( n .eq. PDG_selectron2  ) Mass_MSf(2,2,1) = val
	if( n .eq. PDG_sup1        ) Mass_MSf(1,3,1) = val
	if( n .eq. PDG_sup2        ) Mass_MSf(2,3,1) = val
	if( n .eq. PDG_sdown1      ) Mass_MSf(1,4,1) = val
	if( n .eq. PDG_sdown2      ) Mass_MSf(2,4,1) = val
	if( n .eq. PDG_snu_mu1     ) Mass_MSf(1,1,2) = val
	if( n .eq. PDG_snu_mu2     ) Mass_MSf(2,1,2) = val
	if( n .eq. PDG_smuon1      ) Mass_MSf(1,2,2) = val
	if( n .eq. PDG_smuon2      ) Mass_MSf(2,2,2) = val
	if( n .eq. PDG_scharm1     ) Mass_MSf(1,3,2) = val
	if( n .eq. PDG_scharm2     ) Mass_MSf(2,3,2) = val
	if( n .eq. PDG_sstrange1   ) Mass_MSf(1,4,2) = val
	if( n .eq. PDG_sstrange2   ) Mass_MSf(2,4,2) = val
	if( n .eq. PDG_snu_tau1    ) Mass_MSf(1,1,3) = val
	if( n .eq. PDG_snu_tau2    ) Mass_MSf(2,1,3) = val
	if( n .eq. PDG_stau1       ) Mass_MSf(1,2,3) = val
	if( n .eq. PDG_stau2       ) Mass_MSf(2,2,3) = val
	if( n .eq. PDG_stop1       ) Mass_MSf(1,3,3) = val
	if( n .eq. PDG_stop2       ) Mass_MSf(2,3,3) = val
	if( n .eq. PDG_sbottom1    ) Mass_MSf(1,4,3) = val
	if( n .eq. PDG_sbottom2    ) Mass_MSf(2,4,3) = val
	if( n .eq. PDG_Z           ) Mass_MZ = val
	if( n .eq. PDG_W           ) Mass_MW = val
	if( n .eq. PDG_h0          ) Mass_Mh0 = val
	if( n .eq. PDG_HH          ) Mass_MHH = val
	if( n .eq. PDG_A0          ) Mass_MA0 = val
	if( n .eq. PDG_Hp          ) Mass_MHp = val
	if( n .eq. PDG_neutralino1 ) Mass_MNeu(1) = val
	if( n .eq. PDG_neutralino2 ) Mass_MNeu(2) = val
	if( n .eq. PDG_neutralino3 ) Mass_MNeu(3) = val
	if( n .eq. PDG_neutralino4 ) Mass_MNeu(4) = val
	if( n .eq. PDG_chargino1   ) Mass_MCha(1) = val
	if( n .eq. PDG_chargino2   ) Mass_MCha(2) = val
	if( n .eq. PDG_gluino      ) Mass_MGl = val
	if( n .eq. PDG_gravitino   ) Mass_MGrav = val
	goto 7

* Block NMIX
2002	read(line, *, err=98, end=98) n1, n2, val
	Array(NMix_ZNeu, n1, n2, 4) = val
	goto 7

* Block UMIX
2003	read(line, *, err=98, end=98) n1, n2, val
	Array(UMix_UCha, n1, n2, 2) = val
	goto 7

* Block VMIX
2004	read(line, *, err=98, end=98) n1, n2, val
	Array(VMix_VCha, n1, n2, 2) = val
	goto 7

* Block STAUMIX
2005	read(line, *, err=98, end=98) n1, n2, val
	Array(StauMix_USf, n1, n2, 2) = val
	goto 7

* Block STOPMIX
2006	read(line, *, err=98, end=98) n1, n2, val
	Array(StopMix_USf, n1, n2, 2) = val
	goto 7

* Block SBOTMIX
2007	read(line, *, err=98, end=98) n1, n2, val
	Array(SbotMix_USf, n1, n2, 2) = val
	goto 7

* Block ALPHA
2008	read(line, *, err=98, end=98) Alpha_Alpha
	goto 7

* Block HMIX
2009	read(line, *, err=98, end=98) n, val
	HMix_Q = blockQ
	if( n .eq. 1 ) HMix_MUE = val
	if( n .eq. 2 ) HMix_TB = val
	if( n .eq. 3 ) HMix_VEV = val
	if( n .eq. 4 ) HMix_MA02 = val
	goto 7

* Block GAUGE
2010	read(line, *, err=98, end=98) n, val
	Gauge_Q = blockQ
	if( n .eq. 1 ) Gauge_g1 = val
	if( n .eq. 2 ) Gauge_g2 = val
	if( n .eq. 3 ) Gauge_g3 = val
	goto 7

* Block MSOFT
2011	read(line, *, err=98, end=98) n, val
	MSoft_Q = blockQ
	if( n .eq. 1  ) MSoft_M1 = val
	if( n .eq. 2  ) MSoft_M2 = val
	if( n .eq. 3  ) MSoft_M3 = val
	if( n .eq. 21 ) MSoft_MHd2 = val
	if( n .eq. 22 ) MSoft_MHu2 = val
	if( Range(n, 31, 33) ) MSoft_MSL(n - 30) = val
	if( Range(n, 34, 36) ) MSoft_MSE(n - 33) = val
	if( Range(n, 41, 43) ) MSoft_MSQ(n - 40) = val
	if( Range(n, 44, 46) ) MSoft_MSU(n - 43) = val
	if( Range(n, 47, 49) ) MSoft_MSD(n - 46) = val
	goto 7

* Block AE
2012	read(line, *, err=98, end=98) n1, n2, val
	Ae_Q = blockQ
	if( n1 .eq. 3 .and. n2 .eq. 3 ) Ae_Atau = val
	goto 7

* Block AU
2013	read(line, *, err=98, end=98) n1, n2, val
	Au_Q = blockQ
	if( n1 .eq. 3 .and. n2 .eq. 3 ) Au_At = val
	goto 7

* Block AD
2014	read(line, *, err=98, end=98) n1, n2, val
	Ad_Q = blockQ
	if( n1 .eq. 3 .and. n2 .eq. 3 ) Ad_Ab = val
	goto 7

* Block YE
2015	read(line, *, err=98, end=98) n1, n2, val
	Ye_Q = blockQ
	if( n1 .eq. 3 .and. n2 .eq. 3 ) Ye_Ytau = val
	goto 7

* Block YU
2016	read(line, *, err=98, end=98) n1, n2, val
	Yu_Q = blockQ
	if( n1 .eq. 3 .and. n2 .eq. 3 ) Yu_Yt = val
	goto 7

* Block YD
2017	read(line, *, err=98, end=98) n1, n2, val
	Yd_Q = blockQ
	if( n1 .eq. 3 .and. n2 .eq. 3 ) Yd_Yb = val
	goto 7

* Block SPINFO, DCINFO
2018	goto 7

* Decays
3000	read(line, *, err=98, end=98) val, n, (id(i), i = 1, n)
	if( decay + n + 1 .ge. LengthDecay ) then
	  print 101, "Out of memory in input line", lineno
	  block = -1
	else
	  do i = 1, n
	    Decay(decay) = Decay_Entry(id(i), n + 1)
	    decay = decay + 1
	  enddo
	  Decay(decay) = val
	  decay = decay + 1
	  Decay(parent) = Decay_Entry(parent_id, decay - parent)
	endif
	goto 7

99	close(10)
	end


************************************************************************
	
	integer function SLHATokens(s)
	implicit none
	character*(*) s

	integer si, di, c

	SLHATokens = 0
	di = 0
	do si = 1, len(s)
	  c = ichar(s(si:si))
	  if( c .eq. 35 ) goto 1
	  if( c .eq. 32 .or. c .eq. 9 ) then
	    if( di .ne. 0 .and.
     &          s(di:di) .ne. ' ' .and. s(di:di) .ne. '=' ) then
	      SLHATokens = SLHATokens + 1
	      di = di + 1
	      s(di:di) = ' '
	    endif
	  else
	    if( c .ne. 61 .or. di .eq. 0 .or. s(di:di) .ne. ' ' )
     &        di = di + 1
	    if( c .gt. 96 ) c = iand(c, 223)
	    s(di:di) = char(c)
	  endif
	enddo

1	do si = di + 1, len(s)
	  s(si:si) = ' '
	enddo
	end

